/** TODO: remove `@ts-ignore` */
declare type ActionType = string;
declare type ActionTypesDictionary = {
    [key in ActionType]: true;
};
declare type DependenciesDictionary = {
    [key in NodeId]: Node;
};
declare type NodeId = string;
declare type Node = {
    id: NodeId;
    domain: string;
    name: string;
    actionTypes: ActionTypesDictionary;
    dependencies: DependenciesDictionary;
    stackWorker: (ctx: Ctx) => any;
};
declare const NODE: unique symbol;
declare type ActionCreator<Input = void, Payload = Input> = {
    (input: Input): Action<Payload>;
    getType: () => string;
    [NODE]: Node;
};
declare type Atom<T> = {
    (state: StateBasic, action: Action<any>): StateBasic;
    [NODE]: Node;
};
declare type Unit<T = unknown> = (ActionCreator<any, T>) | (Atom<T>);
declare type StateBasic = {
    [key in string]: {
        [key in string]: any;
    };
};
declare type StackWorker = (ctx: Ctx) => any;
declare type Stack = StackWorker[];
declare type Ctx = {
    state: StateBasic;
    type: string;
    payload: any;
    stack: Stack;
    stateNew: StateBasic;
    statePlain: {
        [key in string]: any;
    };
    visited: {
        [key in string]: true;
    };
    isChanged: boolean;
    write: (node: Node, value: any) => void;
};
declare type Action<Payload, Type extends string = string> = {
    type: Type;
    payload: Payload;
};
export declare function createActionCreator<Input = void, Payload = Input>(type?: string | [string], mapper?: (input: Input) => Payload): ActionCreator<Input, Payload>;
export declare const actionDefault: ActionCreator<void, void>;
export declare function createAtom<State>(name: string | [string, string], initialState: State, handle: (reduce: <T>(dependency: Unit<T>, reducer: (state: State, value: T) => State) => void) => any): Atom<State>;
export declare function createAtom<State>(initialState: State, handle: (reduce: <T>(dependency: Unit<T>, reducer: (state: State, value: T) => State) => void) => any): Atom<State>;
export declare function getState<T>(state: StateBasic, atom: Atom<T>): T | undefined;
export declare function getNode(target: Unit): Node;
export declare function map<T, _T = unknown>(atom: Atom<_T>, mapper: (dependedAtomState: _T) => T): Atom<T>;
export declare function map<T, _T = unknown>(name: string | [string, string], atom: Atom<_T>, mapper: (dependedAtomState: _T) => T): Atom<T>;
export declare function combine<T extends {
    [key in string]: Atom<any>;
} | TupleOfAtoms>(shape: T): Atom<{
    [key in keyof T]: T[key] extends Atom<infer S> ? S : never;
}>;
export declare function combine<T extends {
    [key in string]: Atom<any>;
} | TupleOfAtoms>(name: string, shape: T): Atom<{
    [key in keyof T]: T[key] extends Atom<infer S> ? S : never;
}>;
declare function storeGetState<TargetAtom extends Atom<any>>(target: TargetAtom): TargetAtom extends Atom<infer S> ? S : never;
declare function storeGetState(): StateBasic;
declare function storeSubscribe<TargetAtom extends Atom<any>>(target: TargetAtom, listener: (state: TargetAtom extends Atom<infer S> ? S : never) => any): () => void;
declare function storeSubscribe(listener: (action: Action<any>) => any): () => void;
export declare type Store = {
    dispatch: (action: Action<any>) => void;
    subscribe: typeof storeSubscribe;
    getState: typeof storeGetState;
};
export declare function createStore(atom: Atom<any>, preloadedState?: {}): Store;
declare type TupleOfAtoms = [Atom<any>] | [Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>] | [Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>, Atom<any>];
export {};
